;; -*- Mode: Scheme; tab-width: 4 -*-

;; build an image file from scratch, given the code, and a vector of
;; constants.

(define code-type 'bytecode)

;; utilities
(define (shift-left val n)
  (if (zero? n)
      val
      (shift-left (* val 2) (- n 1))))

(define (shift-right val n)
  (if (zero? n)
      val
      (shift-right (/ val 2) (- n 1))))

(define tag-value
  (let ((tags
		 (map (lambda (x)
				(list (car x)
					  (shift-left (cadr x) (caddr x))))
			  ;; kind        value shift
			  '(;; immediate types
				(CHAR		1  1)
				(BOOL		3  1)
				(NIL		5  1)
				(UNDEFINED	7  1)
				;; pointer types
				(ENV		2  2)
				(CLOSURE	3  2)
				(SAVE		4  2)
				(STRING		5  2)
				(VECTOR		6  2)
				(TUPLE		7  2)
				(CONS		8  2)
				(SYMBOL		9  2)
				(CODE		10 2)
				))))
	(lambda (kind value)
	  (+ (cadr (assq kind tags)) (shift-left value 8)))))

(define (write-bootstrap-file output-port code-string constants)

  (let ((offset 0)
		(symbol-offsets '()))
	
	(define (little-encode-long n)
	  ;; encode an integer as four little-endian bytes
	  (if (< n 0)
		  (set! n (+ #x100000000 n)))
	  (let loop ((i 4)
				 (n n)
				 (r '()))
		(if (zero? i)
			(reverse r)
			(loop (- i 1)
				  (quotient n #x100)
				  (cons (modulo n #x100) r)))))

	(define (emit byte)
	  (write-char (ascii->char byte) output-port)
	  (set! offset (+ offset 1)))

	(define (emit-long n)
	  (for-each emit (little-encode-long n)))

	(define (how-many x n)
	  (quotient (- (+ x n) 1) n))

	;; we return the offset of the object written. for example, when
	;; writing a pair, we write the car, write the cdr, then emit a
	;; cons cell containing the pointers to each, and finally return
	;; the offset of that cons cell.

	(define (emit-object x)
	  ;; immediates: char, bool, nil, undefined
	  ;; tuples: env, closure, save, string, vector, tuple, cons, symbol, code
	  (cond ((integer? x)
			 (+ 1 (shift-left x 1)))
			((char? x)
			 (tag-value 'CHAR (char->ascii x)))
			((boolean? x)
			 (tag-value 'BOOL (if x 1 0)))
			((null? x)
			 (tag-value 'NIL 0))
			((pair? x)
			 (let ((car-offset (emit-object (car x)))
				   (cdr-offset (emit-object (cdr x))))
			   (let ((result offset))
				 (emit-long (tag-value 'CONS 2))
				 (emit-long car-offset)
				 (emit-long cdr-offset)
				 result)))
			((vector? x)
			 (let ((values (map emit-object (vector->list x))))
			   (let ((result offset))
				 (emit-long (tag-value 'VECTOR (vector-length x)))
				 (for-each emit-long values)
				 result)))
			((string? x)
			 (let ((len (string-length x)))
			   (let ((wlen (how-many len 4))
					 (result offset))
				 (emit-long (tag-value 'STRING (+ wlen 1)))
				 ;; emit the string length
				 (emit-long (+ 1 (shift-left len 1)))
				 ;; emit the string
				 ;; XXX Why by char? XXX
				 (for-each
				  (lambda (ch) (emit (char->ascii ch)))
				  (string->list x))
				 ;; pad the string
				 (let loop ((n (- (* wlen 4) len)))
				   (cond ((not (zero? n))
						  (emit 0)
						  (loop (- n 1)))))
				 result)))
			((symbol? x)
			 (let ((probe (assq x symbol-offsets)))
			   (if probe
				   (cdr probe)
				   (let ((string-offset (emit-object (symbol->string x))))
					 (let ((result offset))
					   (emit-long (tag-value 'SYMBOL 1))
					   (emit-long string-offset)
					   (set! symbol-offsets (cons (cons x result) symbol-offsets))
					   result)))))
			(else
			 (error "unsupported object type"))))

	(define (emit-symbol-vector)
	  (let ((result offset))
		(emit-long (tag-value 'VECTOR (length symbol-offsets)))
		(for-each
		 (lambda (entry) (emit-long (cdr entry)))
		 symbol-offsets)
		result))

	;; emit a header
	(for-each
	 (lambda (ch) (write-char ch output-port))
	 (string->list "(lunacy-bootstrap-image)"))
	(write-char (ascii->char 10) output-port)

	;; structure of a bootstrap image:
	;; <space for gc-relocate offset> <constants> ... <constant-vector> <code-object> <closure> <offset to closure>
	(emit-long 1)

	(let ((cv-offset (emit-object (list->vector constants)))
		  (cs-offset (emit-object code-string))
		  (sv-offset (emit-symbol-vector)))
	  ;; --------------------------------------------------
	  ;; emit the code object
	  ;; --------------------------------------------------
	  (let ((code-offset offset))
		(emit-long (tag-value 'CODE 2))
		;; pointer to constant vector
		(emit-long cv-offset)
		;; pointer to code string
		(emit-long cs-offset)
		(let ((closure-offset offset))
		  ;; closure = { tag, code, pc, lenv }
		  (emit-long (tag-value 'CLOSURE 3))
		  ;; pointer to code object
		  (emit-long code-offset)
		  ;; pc (remember, BOX_INTEGER(0) == 1)
		  (emit-long 1)
		  ;; lenv
		  (emit-long sv-offset)
		  ;; *** HACK ALERT ***
		  ;; The environment pointer is deliberately not set to zero due to confusion on
		  ;; the part of gc_relocate, which will think this points back to the beginning of
		  ;; the image/heap.  But we are also abusing the environment pointer as a way of
		  ;; passing in the constant-vector to the bootstrap code.  This can be used to
		  ;; initialize the runtime's symbol table.
		  ;; *** HACK ALERT ***

		  (emit-long closure-offset)
		  ))
	  )
	)
  )


;(define (bootstrap code)
;  ;; for use from another scheme; portable-ish.
;  (let ((tc (transform code (lambda (x) x) #t)))
;	(write tc)
;	(let ((top-level-symbols (cadddr (car tc))))
;	  (if (eq? (car top-level-symbols) 'let-macro)
;		  (set! top-level-symbols (caddr (car (caddr top-level-symbols))));; bad sam
;		  (set! top-level-symbols (cadr (car top-level-symbols))))
;	  ;(display top-level-symbols)
;	  (let ((cc (compile tc '() '())))
;		(let ((vm-pair ((make-bytecoder) cc)))
;		  (let ((cs (car vm-pair))
;				(cv (cdr vm-pair)))
;			(let ((op (open-output-file "test.image")))
;			  (write-bootstrap-file op cs (append cv (list top-level-symbols)))
;			  (close-output-port op)
;			  )))))))
								  
(define (bootstrap code)
  ;; for use from another scheme; portable-ish.
  (let ((tc (transform code (lambda (x) x) #t)))
	(let ((top-level-symbols (cadddr (car tc))))
	  ;;(debug)
	  (if (eq? (car top-level-symbols) 'let-macro)
		  (set! top-level-symbols (caddr (car (caddr top-level-symbols))));; bad sam
		  (set! top-level-symbols (cadr (car top-level-symbols))))
	  ;(display top-level-symbols)
	  (let ((cc (compile tc '() '())))
		(let ((vm-pair ((make-bytecoder) cc)))
		  (let ((cs (car vm-pair))
				(cv (cdr vm-pair)))
			(let ((op (open-output-file "test.image")))
			  (write-bootstrap-file op cs (append cv (list top-level-symbols)))
			  (close-output-port op)
			  )))))))

(define (bootstrap2)
  ;; for use from within lunacy, does timings.
  ;; make sure to do 'make lunacy.scm' first.
  (let ((t0 (%clock))
		(forms (car (read-forms-from-file "lunacy.scm"))))
	(%print "bootstrapping...")
	(let ((t1 (%clock))
		  (tc (transform forms (lambda (x) x) #t)))
	  (let ((top-level-symbols (cadddr (car tc))))
		;; yes, this is horrible. I'm looking directly into the rewritten code
		;; to pull out the top-level-symbols, so we can cheat and use lexref instead
		;; of dynref for known functions.
		(if (eq? (car top-level-symbols) 'let-macro)
			(set! top-level-symbols (caddr (car (caddr top-level-symbols))));; bad sam
			(set! top-level-symbols (cadr (car top-level-symbols))))
		;(display top-level-symbols)
		(let ((t2 (%clock))
			  (cc (compile tc '() '())))
		  (let ((t3 (%clock))
				(vm-pair ((make-bytecoder) cc)))
			(let ((cs (car vm-pair))
				  (cv (cdr vm-pair)))
			  (let ((t4 (%clock))
					(op (open-output-file "test.image")))
				(write-bootstrap-file op cs (append cv (list top-level-symbols)))
				(close-output-port op)
				(let ((t5 (%clock)))
				  (%print
				   `((read ,(%- t1 t0))
					 (transform ,(%- t2 t1))
					 (compile ,(%- t3 t2))
					 (encode ,(%- t4 t3))
					 (write ,(%- t5 t4))
					 (total ,(%- t5 t0))))
				  )))))))))

(define (bootstrap-simple code)
  ;; bootstrap without all the runtime-system hacking
  (let ((tc (transform code (lambda (x) x) #t)))
	(let ((cc (compile tc '() '())))
	  (let ((vm-pair ((make-bytecoder) cc)))
		(let ((cs (car vm-pair))
			  (cv (cdr vm-pair)))
		  (let ((op (open-output-file "test.image")))
			(write-bootstrap-file op cs (append cv (list)))
			(close-output-port op)
			))))))

(define (wrap-file-with-let filename)
  (with-input-from-file filename
	(lambda ()
	  (let loop ((form (read))
				 (forms '()))
		(if (eof-object? form)
			`(let () ,@(reverse forms))
			(loop (read) (cons form forms)))))))

(define (wrap-files-with-let . files)
  (let loop1 ((all-forms '())
			  (files files))
	(if (null? files)
		`(let () ,@all-forms)
		(with-input-from-file (car files)
			(lambda ()
			  (let loop2 ((form (read))
						  (forms '()))
				(if (eof-object? form)
					(loop1
					 (append all-forms (reverse forms))
					 (cdr files))
					(loop2 (read) (cons form forms)))))))))
