;; -*- Mode: Scheme; indent-tabs-mode: nil -*-

(define (make-bytecoder)
  (let ((opcodes
         '((literal             . #\L)
           (constant            . #\C)
           (varref              . #\V)
           (assign              . #\!)
           (save                . #\S)
           (restore             . #\R)
           (save-val            . #\q)
           (test                . #\?)
           (primop              . #\%)
           (invoke              . #\I)
           (jump                . #\J)
           ;;(close             . #\P)
           ;; using newline lets us see line-delimited
           ;; functions in an image file.
           (close               . #\newline)
           (halt                . #\H)
           (new-tuple           . #\T)
           (tuple-store         . #\>)
           (push-env            . #\Z)
           (builtin             . #\B)
           ;; (store0-9         . #\0-#\9)
           ;; (lexref[0-3][0-3] . #\a-p)
           ))
        (primops
         '((%+                  . #\+)
           (%-                  . #\-)
           (%*                  . #\*)
           (%/                  . #\/)
           (%%                  . #\%)
           (%eq?                . #\=)
           (%gt?                . #\>)
           (%lt?                . #\<)
           (%and		. #\&)
           (%or			. #\|)
           (%ge?                . #\})
           (%le?                . #\{)
           (%xor                . #\^)
           (%cons               . #\a)
           (%car                . #\b)
           (%cdr                . #\c)
           (%integer->char      . #\d)
           (%char->integer      . #\e)
           (%apply1             . #\f)
           (%new-tuple          . #\g)
           (%tuple-length       . #\h)
           (%tuple-ref          . #\i)
           (%tuple-set!         . #\j)
           (%build-tuple        . #\k)
           (%undefined          . #\l)
           (%gc                 . #\m)
           (%dump-image         . #\n)
           (%get-typecode       . #\o)
           (%get-environment    . #\p)
           (%room               . #\q)
           (%insns              . #\r)
           (%getcc              . #\s)
           (%putcc              . #\t)
           (%left-shift         . #\u)
           (%right-shift        . #\v)
           (%set-error-handler  . #\w)
           (%get-error-handler  . #\x)
           ))
        (builtins
         '(%print
           %read-char
           %write-char
           %string-compare
           %string-ref
           %string-set!
           %list->string
           %string-append
           %buffer-copy
           %write-string
           %flush
           %assq
           %memq
           %open
           %read
           %write
           %close
           %socket
           %connect
           %bind
           %listen
           %accept
           %clock
           ))
         
        (integer-table "0123456789ABCDEF")
        (lexref-table  "abcdefghijklmnop")
        (new-tuple-table ".;\",")

        )

    (define rest    cdr)
    (define first   car)
    (define second  cadr)
    (define third   caddr)
    (define fourth  cadddr)

    (lambda (code)
      (let ((constants '())
            (n-constants 0))

        (define (new-constant c)

          (define (index c)
            (let loop ((i 1)
                       (l constants))
              (if (null? l)
                  #f
                  (if (equal? c (car l))
                      (- n-constants i)
                      (loop (+ i 1) (cdr l))))))

          (let ((probe (index c)))
            (if probe
                probe
                (begin
                  (set! constants (cons c constants))
                  (set! n-constants (+ n-constants 1))
                  (- n-constants 1)))))

        (define (emit-code code)

          ;; We have two different ways to encode integers in the VM stream,
          ;; this way is more compact but slightly more expensive to unpack.
          ;; It also makes the bytecode more readable, and portable, even to
          ;; 64-bit machines.  The speed difference is only a couple of percent,
          ;; mostly because of all the implied-opcode stuff.

          (define (encode-integer n)
            (let ((neg (negative? n)))
              (let loop ((n (if neg (- 0 n) n))
                         (r (if neg '(#\-) '(#\+))))
                (if (zero? n)
                    r
                    (loop (quotient n #x10)
                          (cons (string-ref integer-table (modulo n #x10)) r))))))
          
          ;; to enable this version instead, uncomment this, comment the above,
          ;; and reverse the sense of the '#if 1' clause near the top of machine.c

          ;(define (encode-integer n)
          ;  (let loop ((n n)
          ;             (i 0)
          ;             (r '()))
          ;    (if (eq? i 4)
          ;        (map ascii->char (reverse r))
          ;        (loop (quotient n 256) (+ i 1) (cons (modulo n 256) r)))))

          (define (encode-literal x)
            ;; should this check be done elsewhere?
            (if (and (pair? x) (eq? (car x) 'quote))
                ;; ok, this is wrong.  it needs to be (quote <literal>)
                (let ((datum (cadr x)))
                  ;; objects that eval to themselves
                  (if (or (number? datum)
                          (string? datum)
                          (char? datum)
                          (boolean? datum))
                      (encode-literal datum)
                      (list->string (cons #\C (encode-integer (new-constant datum))))))
                (let ((num
                       (cond ((integer? x) (+ 1 (* 2 x)))
                             ((boolean? x) (if x #x106 #x06))
                             ((char? x) (+ #x02 (* #x100 (char->ascii x))))
                             ((null? x) #x0a)
                             (else #f))))
                  (if num
                      (list->string (cons #\L (encode-integer num)))
                      (list->string (cons #\C (encode-integer (new-constant x)))))
                  )))

          (define (lookup opcode)
            (cdr (assq opcode opcodes)))
          
          (define (lookup-primop name)
            (assq name primops))

          (define (lookup-builtin name)
            (let loop ((i 0)
                       (l builtins))
              (if (null? l)
                  #f
                  (if (eq? name (car l))
                      i
                      (loop (+ i 1) (cdr l))))))

          (define (emit-insn insn)
            (let ((opcode (car insn)))
              (let ((opchar (lookup opcode)))
                (case opcode
                  ((literal)
                   (encode-literal (second insn)))
                  ((primop)
                   ;; we split the notion of primops and builtins here
                   ;; at the compiler level, we don't see a difference;
                   ;; but at the VM level, we do.
                   (let ((pprobe (lookup-primop (second insn))))
                     (if pprobe
                         (list->string (list opchar (cdr pprobe)))
                         (let ((bprobe (lookup-builtin (second insn))))
                           (if bprobe
                               (list->string (cons #\B (encode-integer bprobe)))
                               (error "unknown primop/builtin")))
                         )))
                  ((new-tuple)
                   (let ((size (second insn)))
                     (if (< size 4)
                         (list->string (list (string-ref new-tuple-table size)))
                         (list->string (cons opchar (encode-integer (second insn)))))))
                  
                  ;; originally, OP_TUPLE_STORE included an index operand
                  ;; since the tuple is always populated in order,
                  ;; we can leave the operand out.  The VM uses the (temporarily unused)
                  ;; 'next' pointer as an index into the 'next' tuple slot to be assigned.
                  ;; this shrinks the bytecode by about 20%, and speeds up the VM.
                  ;; optimized OP_STORE
                  ;;((tuple-store)
                  ;; (string opchar))
                  ;; Ok, that optimization is cool, but it violates the condition
                  ;; that continuations are immutable, and thus breaks call/cc
                  ;; Here's a better version that encodes the argument in the opcode.
                  ((tuple-store)
                   (let ((index (second insn)))
                     (if (< index 10)
                         (number->string index)
                         (list->string (cons opchar (encode-integer (second insn)))))))
                  ((test)
                   (let ((then-code (emit-code (second insn)))
                         (else-code (emit-code (third insn))))
                     (let ((jump-code
                            (list->string
                             (cons (lookup 'jump)
                                   (encode-integer (string-length else-code))))))
                       (string-append
                        (list->string
                         (cons opchar
                               (encode-integer (+ (string-length jump-code) (string-length then-code)))))
                        then-code
                        jump-code
                        else-code
                        ))))
                  ((close)
                   (let ((proc-code (emit-code (second insn)))
                         (name (third insn)))
                     (string-append
                      (list->string (cons opchar (encode-integer (string-length proc-code))))
                      proc-code)))
                  ((save)
                   (let ((proc-code (emit-code (second insn))))
                     (string-append
                      (list->string (cons opchar (encode-integer (string-length proc-code))))
                      proc-code)))
                  ((save-val)
                   (let ((proc-code (emit-code (second insn))))
                     (string-append (string opchar) proc-code)))
                  ((varref assign)
                   (let ((depth (first (second insn)))
                         (index (second (second insn))))
                     (if (and (eq? opcode 'varref) (< depth 4) (< index 4))
                         (list->string (list (string-ref lexref-table (+ (* 4 depth) index))))
                         (list->string
                          (append
                           (cons opchar (encode-integer (first (second insn))))
                           (encode-integer (second (second insn))))))))
                  (else
                   (string opchar))
                  ))))

          (apply string-append (map emit-insn code))
          )
        (cons (emit-code code) (reverse constants))
        ))))

(define (compile-cvm exp cte cts)
  ((make-bytecoder) (compile exp cte cts)))
